# -*- coding: utf-8 -*-
import json
import networkx as nx
import matplotlib.pyplot as plt
import collections
import statistics 
import time
import random

def writePath(G, u, v):
    print("\nHere is the shortest path from", u, "to", v, ":")
    if not u in G or not v in G:
        print("  Error:", u, "and/or", v, "are not in the network")
        return
    try:
        P = nx.shortest_path(G, source=u, target=v)
        for i in range(len(P) - 1):
            t = G.edges[P[i],P[i+1]]["title"]
            print(" ", P[i], "was in", t, "with", P[i+1])
    except nx.NetworkXNoPath:
        print("  No path exists between", u, "and", v)


def doHistogramSummary(X, xlabel, ylabel, title):
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.hist(X, bins=20)
    plt.show()

print("Reading input file...")
Movies = []
with open("data.json", "r", encoding="utf-8") as f:
    for line in f.readlines():
        J = json.loads(line)
        Movies.append(J)


print("\nCalculating number of movies per year...")
C = collections.Counter([d["year"] for d in Movies])
plt.xlabel("Year")
plt.ylabel("Frequency")
plt.title("Number of Movies per Year")
plt.bar(list(C.keys()), list(C.values()))
plt.show()


print("\nCalculating number of movies per cast size...")
C = collections.Counter([len(d["cast"]) for d in Movies])
plt.xlabel("Cast Size")
plt.ylabel("Frequency")
plt.title("Number of Movies per Cast Size")
plt.bar(list(C.keys()), list(C.values()))
plt.show()


print("\nCalculating largest movies by cast size...")
Movies = sorted(Movies, key = lambda i: len(i["cast"]), reverse=True)
for i in range(5):
    print(Movies[i]["title"], "=", len(Movies[i]["cast"]))
   
    
print("\nForming multigraph...")
G = nx.MultiGraph()        
for movie in Movies:
    for i in range(0, len(movie["cast"]) - 1):
        for j in range(i + 1, len(movie["cast"])):
            G.add_edge(movie["cast"][i], movie["cast"][j], title = movie["title"])
print("Number of nodes in this multigraph =", G.number_of_nodes())
print("Number of edges in this multigraph =", G.number_of_edges())


print("\nCalculating films per actor...")
D = {}
for v in G.nodes():
    E = list(G.edges(v, data = True))
    S = set()
    for e in E:
        S.add(e[2]["title"])
    D[v] = S
L = sorted(D.items(), key=lambda item: len(item[1]), reverse=True)
for i in range(5):
    print(L[i][0], ":", len(L[i][1]))  
   
    
print("\nCalculating films per acting partnership...")    
D = {}
for e in G.edges():
    D[e[0] + " and " + e[1]] = G.number_of_edges(e[0], e[1])
L = sorted(D.items(), key=lambda kv: kv[1], reverse=True)
for i in range(5):
    print(L[i][0], ":", L[i][1])
 
    
print("\nConverting to simple graph...")
G = nx.Graph()        
for movie in Movies:
    for i in range(0, len(movie["cast"]) - 1):
        for j in range(i + 1, len(movie["cast"])):
            G.add_edge(movie["cast"][i], movie["cast"][j], title=movie["title"])         
print("Number of nodes in simple graph  =", G.number_of_nodes())
print("Number of edges in simple graph  =", G.number_of_edges())
print("Graph Connected?                 =", nx.is_connected(G))


writePath(G, "Catherine Zeta-Jones", "Jonathan Pryce")
writePath(G, "Homer Simpson", "Neil Armstrong")


print("\nAnalysing components of this graph...")
C = [len(c) for c in sorted(nx.connected_components(G), key=len, reverse=True)]
print("Number of components =", len(C))
print("Component sizes      =", C)


print("\nIsolating the largest component...")
C = max(nx.connected_components(G), key=len)
G = G.subgraph(C)
print("Number of nodes in simple graph =", G.number_of_nodes())
print("Number of edges in simple graph =", G.number_of_edges())
print("Graph Connected?                =", nx.is_connected(G))


print("\nCalculating degree centrality...")
D = nx.degree_centrality(G)
L = sorted(D.items(), key=lambda item: item[1], reverse=True)
for i in range(5):
    print(L[i][0], ":", L[i][1])


# WARNING: The following code block will take up to 12 hours to complete    
print("\nCalculating betweenness centrality (please be patient)...")    
start = time.time()
D = nx.betweenness_centrality(G, k=1000)
end = time.time()
print("Time taken =", end-start, "seconds")
L = sorted(D.items(), key=lambda item: item[1], reverse=True)
for i in range(5):
    print(L[i][0], ":", L[i][1])


# WARNING: The following code block will take up to 12 hours to complete    
print("\nCalculating closeness centrality (please be patient)...")
V = [L[i][0] for i in range(len(L))]
D = {}
start = time.time()
for i in range(1000):
    D[V[i]] = nx.closeness_centrality(G, V[i])
end = time.time()
print("Time taken =", end-start, "seconds")
L = sorted(D.items(), key=lambda item: item[1], reverse=True)
for i in range(5):
    print(L[i][0], ":", L[i][1])


print("\nCalculating closeness centrality of Christopher Lee...")
D = nx.shortest_path_length(G, "Christopher Lee")
print(collections.Counter(D.values()))


# WARNING: The following code block will take up to 12 hours to complete
print("\nCalculating closeness centrality of random sample of actors (please be patient)...")
random.shuffle(V)
D = {}
for i in range(1000):
    D[V[i]] = 1 / nx.closeness_centrality(G, V[i])
L = D.values()
print("Mean =", statistics.mean(L))
print("SD   =", statistics.stdev(L))
doHistogramSummary(L, "Average distance to all Actors", "Frequency", "Closeness centrality distribution")
